import os
import re
from typing import TypeVar, Union

T = TypeVar("T")

class D:
    def __init__(self, data):
        self.data = data


___ = D("")


class DesignConfig:

    def __init__(self, data: dict | os._Environ[str] = os.environ):
        self._data = data

        for attr in dir(self):
            if not attr.startswith('_'):
                setattr(self, attr, self(attr))

    def __call__(self, template: str, default: T = '') -> T | str:
        if type(template) != str:
            return template

        reg = re.compile(r'{([^{}]*)}')
        if not reg.search(template):
            return self._get(template, default)

        template = self._get_format(template)
        if default and reg.search(template):
            return default
        else:
            return template

    def _get(self, key, default: T = '') -> T | str:

        if hasattr(self, key):
            value = getattr(self, key)
            if type(value) == D:
                type_value = type(value.data)

                assert type_value in [str, bool, int, float], \
                    f"D-function only handles atomic values, not {type_value}"

                if key not in self._data:
                    value = value.data
                else:
                    value = self._data[key]

                    if type_value is bool:
                        if type(value) == str:
                            if value.lower() in ['false','no','0']:
                                value = False
                        value = bool(value)
                    elif type_value is int:
                        try:
                            value = int(value)
                        except ValueError:
                            value = 0
                    elif type_value is float:
                        try:
                            value = float(value)
                        except ValueError:
                            value = 0

            if type(value) == str:
                return self._get_format(value)
            else:
                return value # type: ignore

        elif default != '':
            return default
        else:
            return key

    def _get_format(self, template: str, path=()) -> str:
        if len(path) != len(set(path)):
            return template
        reg = re.compile(r'{([^{}]*)}')
        if reg.search(template):
            return template.format(**{key: self._get_format(self._get(key), path+(key,)) for key in reg.findall(template)})
        else:
            return template

    def __getitem__(self, item: str) -> str:
        return self(item)

    def __str__(self):
        return '\n'.join([f'{attr}\t->\t{self._get(attr)}' for attr in dir(self) if not attr.startswith('_')])





import os
from design_config import DesignConfig, D, ___

class Config(DesignConfig):

    """ is D("") """
    PROP_ZERO = ___
    
    """ is 'hello' """
    PROP_ONE = 'Hello'

    """ 'World' if in init_data_dict not PROP_TWO
    As usually init_data_dict is os.environ """
    PROP_TWO = D('World')

    """ 'Hello World!' or 'hello {PROP_TWO}!' for any PROP_TWO """
    PROP_THREE = '{PROP_ONE} {PROP_TWO}!'

    PROP_FIVE = D(55)

    PROP_SIX = D(True)

    PROP_SEVEN = D(7.7)


def test_base():
    """ Config() == Config(os.environ) """
    config = Config({'PROP_ONE': 'Hi', 
    'PROP_TWO': 'Alex', 'PROP_FIVE': '5', 'PROP_SIX': 'FALSE','PROP_SEVEN': '7e7', })

    assert config.PROP_THREE == "Hello Alex!"

    assert hasattr(config, 'PROP_FOUR') == False

    assert config['PROP_THREE'] == "Hello Alex!"
    assert config['{PROP_THREE}!!'] == "Hello Alex!!!"
    assert config['{PROP_FOUR}'] == "PROP_FOUR"

    assert config('{PROP_ONE} Tom') == "Hello Tom"
    assert config('PROP_FOUR', 'London') == "London"
    assert config('PROP_TWO', 'London') == "Alex"
    assert config('{PROP_ZERO}', 'London') == ""
    assert config('{PROP_FOUR}', 'London') == "PROP_FOUR"

    assert config.PROP_FIVE == 5
    assert config.PROP_SIX is False
    assert config.PROP_SEVEN == 7e7

if __name__ == '__main__':
    test_base()